import { Component, OnInit, OnDestroy, HostListener, AfterViewChecked, ComponentFactoryResolver } from '@angular/core';
import { ActivatedRoute, Params } from '@angular/router';
import * as $ from 'jquery';

// Component injection imports
import { ViewContainerRef, ViewChild, ComponentRef } from '@angular/core';
import { SystemJsNgModuleLoader, NgModuleFactory, Compiler, Type, NgModuleRef, ComponentFactory } from '@angular/core';

import 'rxjs/add/operator/switchMap';

import { BackendMessage, LogicalMessage, Map, Field } from './term/message';
import { AddComponents, RebindComponents, WindowComponents, SetCursorPosition } from './term/message';
import { TransactionService } from './term/transaction.service';
import { Data } from './term/term.model';
import { TermService } from './term/term.service';
import { TestingService } from './term/testing.service';
import { AppService } from './app.service';

import { AppConfigurationMessage } from './configuration-message.module';
import { ModalService } from './modal.service';

// Component injection imports

/** Simplify template expression */
class FactoryTuple {
    constructor(
        public modFac: NgModuleFactory<any>,
        public cmpFactory: ComponentFactory<any>,
        public componentName: string) { }
}

/** Abstract business component receiving, displaying and sending data */
export abstract class AbstTermComponent implements OnInit, OnDestroy, AfterViewChecked {

    // Component injection location
    @ViewChild('dynamicTarget', { read: ViewContainerRef })
    private dynamicTarget : ViewContainerRef;

    // Keep track of injected component(s)
    private injectedComponents: ComponentRef<any>[] = [];

    // Keep track of injected component(s), by name
    private componentsByName: { [key: string]: ComponentRef<any> } = {}

    // Keep track of injected fields(s), by position
    private fieldsByPosition: { [key: number]: Data } = {}

    // Keep track of message treated
    private messages: LogicalMessage[] = []

    // The next transaction to run from this terminal (CICS "TRANSID" parameter)
    private nextTransactionId: string;

    // Temporary; remove when the new style will be validated (V7-316)
    useModernLegacyStyle: boolean;

    // A message to display in the page footer
    private footerMessage: string;

    // Used by Setup menu. Enforce "Fixed Width" as the defaut setup
    private containerClasses = ['container', 'rd_fixed_width'];

    /** The global configuration */
    public configuration: AppConfigurationMessage;

    /** Current year */
    public currentYear:number;

    /** Parent component */
    private parentComponent: AbstTermComponent;

    constructor(
        public modalService: ModalService,
        private transactionService: TransactionService,
        private route: ActivatedRoute,
        private loader: SystemJsNgModuleLoader,
        private compiler: Compiler,
        private service: TermService,
        private componentFactoryResolver: ComponentFactoryResolver,
        private appService: AppService,
        private testingService: TestingService
    ) {
        service.onReceiveMessageToTerm = this.onReceivedMessage.bind(this);
        this.configuration = transactionService.configuration;
        this.currentYear = Date.now();
        console.log('useModernLegacyStyle: ' + this.useModernLegacyStyle);
   
    }

    
    abstract getMapComponent(mapName : string): any;
    abstract getIdAnchor(): any;
    abstract decodeFieldsSubComponent(): boolean;
    
    protected postAction() {
        // Do Nothing
    }

    getComponent(name : string) {
        return this.componentsByName[name]
    }

    getFieldByPosition(position : number) {
        return this.fieldsByPosition[position]
    }

    setParent(parent: AbstTermComponent) {
        if (parent !== this) {
            this.parentComponent = parent;
        }
    }

    getParent(): AbstTermComponent {
        return this.parentComponent
    }

    ngOnInit(): void {
        let toRun = (params: Params) => this.onInvoke(params);
        let observable = this.route.params.switchMap(toRun);
        observable.subscribe((message: BackendMessage) => this.onReceivedMessage(message));
        // Default theme
        document.body.className = 'rdt_dark';
    }

    ngOnDestroy(): void {
        this.removeAnyExistingComponent();
    }


    public ngAfterViewChecked(): void {

        // Change DOM pointer-events attribute value to enable targetting.
        if (this.haveToChangeTargetAccessibility()) {

            // Disable pointer-event on all 
            this.getIdAnchor().attr("style", "pointer-events: none");

            // Keep accessible as target usefull fields
            let msgs : any[] = [];
            let numberLine : number = 0;
            this.getIdAnchor().children().each(
                function(index, values) {
                    msgs.push(values);
                    numberLine = Math.max(numberLine, $(values).children().children().length);
                }
            )

            // For each line
            for (let i = 0; i<numberLine; i++) {

                let retainDivLine;
                for (let j = 0; j<msgs.length; j++) {
                    let divLine = $(msgs[j]).children().children()[i];
                    if ($(divLine).children().filter(
                        function(index,value) {
                            return !value.classList.contains("lgr_00")
                        }
                    ).length > 0) {
                        retainDivLine = divLine;
                    }
                }


                // Set the last no empty div line attribute pointer-events value to visible
                if (retainDivLine != null) {
                    $(retainDivLine).attr("style", "pointer-events: visible");
                }
            }
        } else {
            // Re-establish default behaviour
            this.getIdAnchor().removeAttr("style");
        }
 
    }

    // Check if condition have to be changed to access target 
    protected haveToChangeTargetAccessibility(): boolean {

        // If there are several message and are in absolute position
        if (this.getIdAnchor().children().length > 1 && this.getIdAnchor().children().children().filter(
            function(index, value) {
                return $(value).css("position") === "absolute"
            }
        ).length > 0 ) {
            return true;
        }

        return false;
    }

    private toggleOverwriteMode(): void {
        this.appService.insertMode = !this.appService.insertMode;
    }

    private changeSetup(event: MouseEvent): void {

        // Ensure element has at least the "container" class
        this.containerClasses = ['container'];

        // Add the class specified by the data-csw attribute, if present
        let target = (<Element>event.target)
        if (target.hasAttribute('data-csw')) {
            this.containerClasses.push(target.getAttribute('data-csw'));
        }
    }

    private changeTheme(event: any): void {

        // Set the class of "body" to the one defined by the data-csw attribute
        let target = (<Element>event.target);
        let theme = target.getAttribute('data-csw');
        document.body.className = theme;
    }

    private onInvoke(params: Params): Promise<BackendMessage> {
        const startMessage: any = { 'resetTransactionData': true };

        const transid: string = params['transid'];
        if (transid.toLowerCase().startsWith('/testmap ')) {
            const args = transid.substring(transid.indexOf(' ') + 1);
            return Promise.resolve(this.testingService.buildTestMessage(args));
        }

        return this.transactionService.runTransaction(transid, startMessage);
    }

    private removeAnyExistingComponent(): void {
        for (let i = 0; i < this.injectedComponents.length; i++) {
            this.injectedComponents[i].destroy();
        }
        this.injectedComponents = [];
        this.componentsByName = {};
        this.fieldsByPosition = {};

        
    }

    private setCursorPosition(cursor: number): void {
    	let injectedField: any = this.getFieldByPosition(cursor);
    	injectedField.initialCursor = true;
    }

    public onReceivedMessage(backendMessage: BackendMessage): void {
        console.log('JSON message received from backend', backendMessage)

        if (backendMessage.error !== undefined) {
            console.log('An error occured on backend: ' + backendMessage.error);
            this.setFooterMessage(backendMessage.error);
            return;
        }

        this.nextTransactionId = backendMessage.nextTransID;

        if (this.configuration.verbose) {
            this.setFooterMessage(backendMessage.serverDescription);
        } else {
        	// Remove potential obsolete error message
            this.setFooterMessage('');
        }

        if (backendMessage.messages == undefined || backendMessage.messages.length == 0) {
            console.log('No backend messages to process');
        } else {
            // Ensure each message is totally processed before the next one is (V7-3021)
            this.processSequentially(backendMessage.messages, 0);
            // backendMessage.messages.forEach(element => {
            //    this.processLogicalMessage(element);
            //});
        }
    }

    /** Set the footer message, or clear it is string is empty/undefined */
    private setFooterMessage(message: string):void {
        if (message == undefined) {
            message = '';
        }

        // Flashing footer if message has changed
        if (message !== '' && message != this.footerMessage) {
            this.animate();
        }

        this.footerMessage = message;
    }

    /** Active footer animation */
    private animate():void {
        $("#idFooter").removeClass("run-animation");
        setTimeout(() => {
            $("#idFooter").addClass("run-animation");
        }, 10);
    }

    /** Process logical messages sequentially (only one usually, but see V7-145) */
    private processSequentially(toProcess: LogicalMessage[], idx: number): void {
        if (idx >= toProcess.length) {
            return;
        }

        // Could use Array.shift, but it messes with debugging console (array is emptied)
        this.processLogicalMessage(toProcess[idx]).then(() => this.processSequentially(toProcess, idx + 1));

    }

    /** Process one logical message, and return a Promise to synchronize on */
    protected processLogicalMessage(message: LogicalMessage): Promise<void> {

        if (this.messages.filter(msg => msg === message).length == 0) {
            
            this.messages.push(message);
            switch (message.command) {
                case 'addComponents':
                    return this.loadAndInsertComponents((message as AddComponents).maps);
                case 'rebindComponents':
                    this.rebindExistingComponents((message as RebindComponents).maps);
                    return Promise.resolve();
                case 'removeAllUnprotected':
                case 'removeAllComponents':
                    this.removeAnyExistingComponent();
                    return Promise.resolve();
                case 'setCursorPosition':
                    this.setCursorPosition((message as SetCursorPosition).cursor);
                    return Promise.resolve();
                case 'windowComponents':
                    return this.loadAndActiveWindowComponents(message as WindowComponents);
                default:
                    let errorMessage = 'Unhandled backend command "' + message.command + '"';
                    console.log(errorMessage); // Ensure this makes its way to the logs
                    throw new Error(errorMessage);
            }
        }
    }

    /** Bind fields values and attributes to existing components */
    private rebindExistingComponents(maps: Map[]): void {

        for (let i = 0; i < maps.length; i++) {
            let map = maps[i];

            // Find the existing component for this map
            let component: ComponentRef<any> = this.componentsByName[map.component];
            if (component === undefined) {
                throw new Error('Component not found in existing ones, cannot rebind: ' + map.component);
            }

            this.decodeFields(component.instance, map);

            /** Clone fields object to trigger method "set data( data: Data )" from dynamic field component */
	        let fields : Field[] = map.fields; 
	        for (let i = 0; i < fields.length; i++) {
	            let field: Field = fields[i];
                let data: Data = component.instance[field.id];
                if (!Array.isArray(data)){ // TODO: Treat array ?
                    let cloneData = new Data ;
                    cloneData.value = data.value;
                    cloneData.initialValue = data.initialValue;
                    cloneData.forceModified = data.forceModified;
                    cloneData.consultMode = data.consultMode;
                    cloneData.attributes = data.attributes;
                    cloneData.initialCursor = data.initialCursor;
                    let pos: number = data.attributes.line * 80 + data.attributes.column;
                    this.fieldsByPosition[pos] = cloneData;
                    component.instance[field.id] = cloneData;
                }
			}

			// If the component contains a subfile, update its record number so the correct page will be displayed
            if(component.instance.isSubfileControl == true){
                component.instance.updateSubfile();
            }
        }
    }

    /** Load components, then insert and bind them */
    private loadAndInsertComponents(maps: Map[]): Promise<void> {

        let proms = this.loadComponents(maps);

        // Now sequentially insert components (order matters)
        // TODO More complex layout may be needed here (see "Cumulative output - the ACCUM option" chapter)
        return Promise.all(proms).then((cmpFactories: Array<ComponentFactory<{}>>) => {
            for (let i = 0; i < maps.length; i++) {

                // Build component, then bind message to UI
                let component = this.dynamicTarget.createComponent(cmpFactories[i]);
                this.injectedComponents.push(component);
                this.componentsByName[maps[i].component] = component;
                this.decodeFields(component.instance, maps[i]);
            }
        });
    }

    /** Set up and return components loading/building factories */
    private loadComponents(maps: Map[]): Array<Promise<ComponentFactory<{}>>> {

		this.setFMTMapWidth(maps);

        // First load and setup asynchronously modules defining map components
        let proms: Array<Promise<ComponentFactory<{}>>> = [];
        for (let i = 0; i < maps.length; i++) {
            let map = maps[i];

            // Search for the angular type corresponding to this component name
            let componentType = this.getMapComponent(map.component);
            if (componentType == null) {
                throw new Error('Unknown component: "' + map.component + '"');
            }

            // Load component-owning module dynamically
            let componentFactory = this.componentFactoryResolver.resolveComponentFactory(componentType);
            proms.push(Promise.resolve(componentFactory));
        }

        return proms;
    }
    
    /* Particular treatment for MFS map:
       Get the map width from the message and set the terminal width
       TODO: Generate this method only for IMS project */
    private setFMTMapWidth(maps: Map[]) {
        // Map shall be unique
        if(maps.length == 1){
            this.containerClasses = ['container'];
            if(maps[0].mapWidth === 132){
                this.containerClasses.push('rd_fixed_width_A7');
            } else {
                // Default width (80)
                this.containerClasses.push('rd_fixed_width');
            }
        }
    }

    /** Load components, then insert and bind them */
    protected loadAndActiveWindowComponents(message: WindowComponents) {

        let maps = message.maps;
        let proms = this.loadComponents(maps);
        
        // Now sequentially insert components (order matters)
        return Promise.all(proms).then((cmpFactories: Array<ComponentFactory<{}>>) => {
            for (let i = 0; i < maps.length; i++) {

                console.log(this);
                // Build component, then bind message to UI
                let component = this.dynamicTarget.createComponent(cmpFactories[i]);

                this.injectedComponents.push(component);
                this.componentsByName[maps[i].component] = component;

                if (this.decodeFieldsSubComponent()) {
                    this.decodeFields(component.instance, maps[i]);
                }

                this.modalService.activeModal(maps[i].component).next({message : message, parentComponent : this});
                
            }
        })

        ;

    }

    private onModuleLoaded(modFac: NgModuleFactory<any>, componentName: string): Promise<FactoryTuple> {
        let moduleType: Type<any> = modFac.moduleType;

        // Use the Compiler to resolve the module's embedded component(s)
        return this.compiler.compileModuleAndAllComponentsAsync<any>(modFac.moduleType)
            .then((factory) => {
                // Return the proper component (others may have been loaded due to SharedModule)
                for (let i = 0; i < factory.componentFactories.length; i++) {
                    if (factory.componentFactories[i].componentType.name === 'MapComponent') {
                        return new FactoryTuple(modFac, factory.componentFactories[i], componentName);
                    }
                }

                throw new Error('MapComponent not found in factory');
            });
    }

    /** Create the component and insert it at the end of "dynamicTarget" */
    private createAndInsertComponent(factory: FactoryTuple): ComponentRef<any> {

        // Instantiate the Module so we can use it as the provider for the new component
        let modRef: NgModuleRef<any> = factory.modFac.create(this.dynamicTarget.parentInjector);

        // Create component and insert it at the end of "dynamicTarget"
        let insertionPos = this.injectedComponents.length;
        let injectedComponent = this.dynamicTarget.createComponent(factory.cmpFactory, insertionPos, modRef.injector);
        this.injectedComponents.push(injectedComponent);
        this.componentsByName[factory.componentName] = injectedComponent;
        return injectedComponent;
    }

    /** Initialize a map component with JSON informations provided by the backend */
    private decodeFields(map: any, inputMap: any): void {

        /** Map technicals fields */
        this.mapFieldToData( map, 'overlay', inputMap.overlay);
        this.mapFieldToData( map, 'startLineNumber',  inputMap.startLineNumber);

        /** Map to business fields */
        let fields : Field[] = inputMap.fields; 
        for (let i = 0; i < fields.length; i++) {
            let field: Field = fields[i];
            let data: Data = map[field.id];
            if (data == null) {
                throw new Error('No entry in map for field ' + field.id)
            }

            // Reflexively set field value
            if (field.data !== undefined) {
                data.value = field.data;
            } else if (field.fields !== undefined && Array.isArray(data)) {

                for (let subMap of field.fields) {
                    let arrayItem : any ={};
                    let index=0;
                    for (let subField of subMap.fields) {
                        let subData = new Data ;
                        subData.value = subField.data;
                        this.fillData(subData, subField);
                        arrayItem[subField.id] = subData;
                    }
                    arrayItem["FIELDS"] = subMap.fields;
                    data.push(arrayItem);
                   
                    index++;
                }
            }

            // Ensure attributes are set on component
            this.fillData(data, field);
        }
    }

    /* Fill data from field. */
    private fillData(data : Data, field: Field): void {

        // Keep track of initial value (for "modified" detection purpose)
        data.initialValue = data.value;

        if (field.attributes !== undefined) {
            // Provide attributes for decoding by the dynamic-field component
            data.attributes = field.attributes;

            // Signal that field data must be sent inconditionnaly (even if not modified)
            // Keep existing value if we did not receive attributes (V7-139, V7-127)
            data.forceModified = field.attributes.forceModified;

            // Update consultation mode (readOnly & disabled)
            data.consultMode = field.attributes && (field.attributes.protection === 'PROT' || field.attributes.protection === 'ASKIP');
            
            let pos: number = data.attributes.line * 80 + data.attributes.column;
            this.fieldsByPosition[pos] = data;
        }

        if (field.initialCursor !== undefined) {
            data.initialCursor = field.initialCursor;
        }
    }

    /* Map field to data value. */
    private mapFieldToData(map : any, data: string, field: any): void {

        if (map[data] !== undefined && field !== undefined) {
            map[data] = field;
        }
    }

    /** React to attention keys */
    @HostListener('document:keydown', ['$event'])
    onKeyDownListener(e: KeyboardEvent) {
    	let activeElementDoc : Element = document.activeElement;
        let activeElement: string = activeElementDoc.id;
        
        let attentionKey: string = null;

        // Robustness switch from key / which / keyCode
        // See https://keycode.info/ for key codes
        let key = e.key || '';
        let keyCode = e.which || e.keyCode || -1;
        let groups = [];
        if (keyCode == 13 || key === 'Enter') {
            attentionKey = 'ENTER';
        }  else if (this.transactionService.is5250() && (keyCode == 33 || key === 'PageUp')) {
        	// Update the subfile control page. if it can't be udpated, set attentionKey to send a message to the backend
            if(!this.processPaging(e)){
                attentionKey = 'PAGEUP';
            }
        }  else if (this.transactionService.is5250() && (keyCode == 34 || key === 'PageDown')) {
        	// Update the subfile control page. if it can't be udpated, set attentionKey to send a message to the backend
            if(!this.processPaging(e)){
                attentionKey = 'PAGEDOWN';
            }
        } else if ((keyCode == 112 && e.ctrlKey) || (keyCode == 45 && e.altKey)) {
            // Ctrl + F1 or Alt + Insert for PA1 (V7-4104)
            attentionKey = 'PA1';
        } else if ((keyCode == 113 && e.ctrlKey) || (keyCode == 36 && e.altKey)) {
            // Ctrl + F2 or Alt + Home for PA2 (V7-4104)
            attentionKey = 'PA2';
        } else if ((keyCode == 114 && e.ctrlKey) || (keyCode == 33 && e.shiftKey)) {
            // Ctrl + F3 or Shift + PageUp for PA3 (V7-4104)
            attentionKey = 'PA3';
        }  else if (keyCode >= 112 && keyCode <= 123) {
			let pfk = (keyCode - 111);
			if (e.shiftKey) {
				// Shift + Fx for PF13 to PF24 (V7-2657)
				pfk += 12;
			}
            attentionKey = 'PF' + pfk;
        } else if (groups = key.match(/^F(\d+)$/)) {
            attentionKey = 'PF' + groups[1];
        } else if (keyCode == 19) {
            // "Pause" key on a PC keyboard
            attentionKey = 'CLEAR';
        }

        if (attentionKey !== null) {
            console.log('Attention key pressed (' + attentionKey + ')');
            e.stopImmediatePropagation();
            e.preventDefault();

            // Build a message to the backend
            let termMessage: any = { 'attentionKey': attentionKey, 'activeField': activeElement, 'fields': [] };
            for (var m = 0; m < this.injectedComponents.length; m++) {
                let injectedComponent: any = this.injectedComponents[m];
                let componentName: string = this.getComponentName(injectedComponent);
                let map: any = injectedComponent.instance;

                // Collect fields value and their "modified" status
                let FIELDS: string[] = map.FIELDS;
                for (var i = 0; i < FIELDS.length; i++) {
                    let fieldName: string = FIELDS[i];
                    let data: Data = map[fieldName];
                    if (Array.isArray(data)) {
                   		let index=0;
                    	for (var a = 0; a < data.length; a++) {
                    		let arrayData = data[a];
                    		for (let subFieldName of arrayData.FIELDS) {
                    			let subField: Data = arrayData[subFieldName.id];
	                    		let modified: boolean = subField.value != subField.initialValue;
		
			                    // Field value is only sent if modified (V7-130)
			                    if (modified || subField.forceModified) {
			                        let fieldState: any = { 'id': subFieldName.id + "_" + index, 'value': subField.value };
			                        termMessage.fields.push(fieldState);
			                    }
			                }
                    		index++;
                    	}
                    } else {
	                    let modified: boolean = data.value != data.initialValue;
	
	                    // Field value is only sent if modified (V7-130)
	                    if (modified || data.forceModified) {
	                        let fieldState: any = { 'id': fieldName, 'value': data.value, ...(data.attributes.isPassword && {'ispassword': data.attributes.isPassword})};
	                        termMessage.fields.push(fieldState);
	                    }
	                }
                }
            }

            this.postAction();

            // Send message to frontend; wait for an answer to react on
            this.transactionService.runTransaction(this.nextTransactionId, termMessage)
                .then((message: BackendMessage) => this.onReceivedMessage(message))//.then(() => console.log("apres receive"))
                .catch((ex) => {
                    console.error('Error invoking next transaction', ex);
                });
        }
    }

    private processPaging(key: KeyboardEvent){
        for (var m = 0; m < this.injectedComponents.length; m++) {
            let injectedComponent: any = this.injectedComponents[m];
            let map: any = injectedComponent.instance;
            if(map.isSubfileControl == true){
                return map.doPaging(key);
            }
        }
        return false;
    }

    /** Get component name */
    private getComponentName(injectedComponent: ComponentRef<any>): string {
		let keys: any = Object.keys(this.componentsByName);
        for (let key of keys) {
        	let component: ComponentRef<any> = this.componentsByName[key];
            if (component === injectedComponent) {
            	return key;
            }
        }
        return "";
    }
}
