import { Component,  Input, SystemJsNgModuleLoader, Compiler, ComponentFactoryResolver } from '@angular/core';
import { NgModule } from '@angular/core';
import { ModalService, MODALSERVICE } from './modal.service';
import { Subscription } from 'rxjs';
import { WindowComponents, BackendMessage } from './term/message';
import { AbstTermComponent } from './abs-term.component';
import { mapSubComponents } from './maps/mapComponents';
import { TransactionService } from './term/transaction.service';
import { ActivatedRoute } from '@angular/router';
import { TermService } from './term/term.service';
import { TestingService } from './term/testing.service';
import { CommonsModule } from './commons.module';
import { AppService } from './app.service';

const KEY_ESC = 27;
declare var $: any;

@Component({
    selector: 'modal-confirm',
    template: `
        <div [id]="name" class="modal" tabindex="-1" role="dialog">
            <div class="modal-dialog" style="position: absolute">
                <div class="modal-content" [ngStyle]="positionStyle()" [style.border-color]="borderColor">
                    <div *ngIf="title" class="modal-header" >
                        <h5>{{title}}</h5>
                    </div>
                    <div class="modal-body" >
                        <div #dynamicTarget></div>
                    </div>
                </div>
            </div>
        </div>
        `
})

/** Component Modal */
export class ModalComponent extends AbstTermComponent {

    constructor(
        modalService: ModalService,
        transactionService: TransactionService,
        route: ActivatedRoute,
        loader: SystemJsNgModuleLoader,
        compiler: Compiler,
        service: TermService,
        componentFactoryResolver: ComponentFactoryResolver,
        appService: AppService,
        testingService: TestingService
    ) {
        super(modalService, transactionService, route, loader, compiler, service, componentFactoryResolver, appService, testingService);
    }

    @Input() name: string;

    // Position : upper-left corner 
	/** The left position. */
    private positionLeft: number;
	/** The top position. */
	private positionTop: number;
	
	/** The cursor position from which displaying modal. Actually not exploited.*/
	private cursorPosition: boolean;
	
	/** The size's height. Actually not exploited, modal height is set by the content */
	private height: number;
	
	/** The size's width. */
    private width: number;

    /** Color border */
    public borderColor: string;

    /** Title */
    public title: string;

    private subscription: Subscription;

    getMapComponent(mapName: string) {
        return mapSubComponents[mapName];
    }

    getIdAnchor() : any  {
        return $("#" + this.name).children().children().children();
    }

    protected haveToChangeTargetAccessibility(): boolean {
        return true;
    }

    decodeFieldsSubComponent(): boolean {
        return true;
    }

    postAction() {
         this._hideDialog();
    }
    
    public onReceivedMessage(backendMessage: BackendMessage): void {
        // Destroy current modal component
        this.getComponent(this.name).destroy();
        // Transfer to main component receiver
        this.getParent().onReceivedMessage(backendMessage);
    }

    private _defaults = {
        title: 'Confirmation',
        message: 'Do you want to cancel your changes?',
        cancelText: 'Cancel',
        okText: 'Yes'
    };

    activate(): Promise<boolean> {
        console.log('Displays modal ' + this.name );
        let promise = new Promise<boolean>(resolve => {
            this._show(resolve);
        });

        return promise;
    }

    private _show(resolve: (boolean) => any) {
        document.onkeyup = null;
        setTimeout(() => $('#' + this.name).modal('show'), 50);
        this.modalService.pushModal(this.name);
        document.onkeyup = (e: any) => {
            if (e.which == KEY_ESC) {
                this._hideDialog();
                return resolve(false);
            }
        };
    }

    private _hideDialog() {
        window.setTimeout(() => $('#' + this.name).modal('hide'), 0);
        this.modalService.popModal(this.name);
    }

    ngOnInit(): any {
        this.subscription = this.modalService.activeModal(this.name).subscribe( ({message, parentComponent}) => {
            this.activate();
            this.setParent(parentComponent);
            this.processLogicalMessage(message);
            this.mapWindowAttributes(message as WindowComponents);
        });
    }

    ngOnDestroy(): void {
        if (this.subscription) {
            this.subscription.unsubscribe();
        }
    }

    private mapWindowAttributes(message: WindowComponents) {
        
        message.maps.forEach(windowMap => {
            if (windowMap.borderColor) {
                this.borderColor = windowMap.borderColor;
            }

            if (windowMap.title) {
                this.title = windowMap.title;
            }

            if (windowMap.positionLeft) {
                this.positionLeft = windowMap.positionLeft;
            }

            if (windowMap.positionTop) {
                this.positionTop = windowMap.positionTop;
            }

            if (windowMap.height) {
                this.height = windowMap.height;
            }

            if (windowMap.width) {
                this.width = windowMap.width;
            }

            if (windowMap.cursorPosition) {
                this.cursorPosition = windowMap.cursorPosition;
            }

        });
    }


    /**  Set position as upper-left corner */
    positionStyle() {

        /** Constante to retrieve instead */
        // lrg height css
        let baseHeight = 1.4;
        // Pixel
        let sizePanel = 720;
        // Base size in pixel
        let baseSize = 16;

        let colWidthPx = sizePanel / 80;

        let style = "{";

        if (this.positionLeft && this.positionTop) {
            let newLeftPosition = (((this.positionLeft) * colWidthPx ) / baseSize);
            style += '"left": "' + newLeftPosition + 'em"';
            let newTopPosition = (((this.positionTop) * baseHeight));
            style += ',"top": "' + newTopPosition + 'em"';
        }

        if (this.width && this.height) {
            let width = (((this.width +  2 /** RPG add border */) * colWidthPx) / baseSize);
            style += ',"width":"'+ width + 'em"';

            let height = (this.height *  baseHeight);
            style += ',"height":"'+ height + 'em"'; 
        }

        style += "}";
       
        return JSON.parse(style);
    }
    
}

@NgModule({
    imports: [
        CommonsModule
    ],
    declarations: [ModalComponent],
    providers: [MODALSERVICE],
    exports: [ModalComponent]
})
export class ModalModule {

}
